// This program reads out PWM Signals from a RC Receiver. 
// The signals can be remaped to get a custom slope signal
// The signal can be set onto a monster motor shield system to control any DC device

// ESP 32 Board eg. Wemos D1
// Monster Motor Shield


#define ESP32
//#define ESP8266

//Monster Motor
// INA1 D7
// INB1 D8
// PWM1 D5

// INA2 D4
// INB2 D9
// PWM2 D6


// NodeMCU    ESP32 R32
#define D0    16  // M1 Cw
#define D1    5   // M1 CCw  SCL
#define D2    26   // PWM1    SDA
#define D3    25   //       –flash kann zu problemen führen
#define D4    17   // M2 CCw 
#define D5    16
#define D6    27  // M2 CCw 
#define D7    14  // PWM2
#define D8    12

#define D9    13
#define D10   5
#define D11   23  // Ext_RC_IR_1  ok 
#define D12   19  // Ext_RC_IR_2  ok
#define D13   18  // Ext_RC_IR_3  ok



#define CHANNEL_1_PIN D11   //D17  // 13 CH3 D13 OK
#define CHANNEL_2_PIN D12   //D8   // 15 CH2 // Ext_RC_IR_2  D12 ok
#define CHANNEL_3_PIN D13   //D5   // 14 CH1 // Ext_RC_IR_3  D11ok

#define M1_PWM_Pin D5       // 13 ///M1 MMP: D5 ok
#define M1_PWM_Channel 0    //D5 //0// //  ESP32   ledcAttachPin(LED_PIN, LEDC_CHANNEL_0);
#define M1_CW_Pin  D8       // 16    //M1 MMP: D7 ok
#define M1_CCW_Pin D7       //  5  M1 MMP: D8 ok

#define M2_PWM_Pin D6       //18  4   //  M2 MMP: D6 ok
#define M2_PWM_Channel 1    //D6//1 //  ESP32   ledcAttachPin(LED_PIN, LEDC_CHANNEL_0);
#define M2_CW_Pin  D9       //M2 MMP: D9 ok
#define M2_CCW_Pin D4       //M2 MMP: D4 ok

// use 13 bit precission for LEDC timer
#define LEDC_TIMER_13_BIT  13
// use 5000 Hz as a LEDC base frequency
#define LEDC_BASE_FREQ     5000

void ledcAnalogWrite(uint8_t channel, uint32_t value, uint32_t valueMax = 255) {

  // calculate duty, 8191 from 2 ^ 13 - 1
  uint32_t duty = (8191 / valueMax) * min(value, valueMax);


#ifdef ESP32
  // write duty to LEDC
  //  ledcWrite(0, 2);
#endif
}

// ########  RC_INTERFACE  ######################################################################
// read out PWM Signals from receiver
class RC_INTERFACE {
  public:
    void isr_setup(int, int);
    void calcSignal(void);
    int ramp_signal(void);
    int get_cur_rc_value(void);
    int get_pulse_time(void);

    volatile unsigned long  timer_start;
    volatile int last_interrupt_time;

    int _pin;

  private:
    int _timer_range;
    int cur_isr_value = 0;
    volatile int long pulse_time = 0;
};
int RC_INTERFACE::get_pulse_time() {

  return pulse_time;
}
void RC_INTERFACE::isr_setup(int isr_pin, int timer_range) {
  _timer_range = timer_range;
  timer_start = 1000;
  _pin = isr_pin;
}
void RC_INTERFACE::calcSignal() {
  int test = digitalRead(_pin);

  last_interrupt_time = micros();

  if (digitalRead(_pin) == HIGH) {
    timer_start = micros();
  }

  else {
    if (timer_start != 0) {
      /* pulse_time =  micros();*/
      pulse_time = ((volatile int)micros() - timer_start);
      timer_start = 0;
    }
  }

}
int RC_INTERFACE::ramp_signal() {
  int val = 0;

  val = (pulse_time - _timer_range) / 5;
  int max_val = 100;
  if (val > max_val)val = max_val;
  if (val < -max_val)val = -max_val;
  cur_isr_value = val;

  return val;
}
int RC_INTERFACE::get_cur_rc_value() {
  return cur_isr_value;
}

RC_INTERFACE RC1, RC2, RC3;

// ################# Monster Motor Shield #########################################################
// control Monster Motor Shield 
class MOTOR_CONTROL {
  public:
    void setup(int, int, int, float, int, int);
    int PWM_from_Ramp(int, int, int);
    void GO(int, int);
    float Get_Cur_PWMval(void);
    void setup_pwm_ESP32(int, int, int, int);
    void PWM(int);

  private:
    int _PWM_Pin;
    int _PWM_CHannel;
    int _ledc_timer_res;
    int _ledc_base_freq;
    int _CW_Pin;
    int _CCW_Pin;
    float _ramp_scale;
    int _pwm_max;
    int _mid_pos;
    float _curr_pwm_val;
};
void MOTOR_CONTROL::setup_pwm_ESP32(int PWM_Pin, int pwm_ch, int ledc_timer_res, int ledc_base_freq) {

  _PWM_CHannel = pwm_ch;
  _PWM_Pin = PWM_Pin;
  _ledc_timer_res = ledc_timer_res;
  _ledc_base_freq = ledc_base_freq;

#ifdef ESP32
  ledcSetup(_PWM_CHannel, _ledc_base_freq, _ledc_timer_res);
  ledcAttachPin(_PWM_Pin, _PWM_CHannel);
#endif

}
void MOTOR_CONTROL::setup(int PWM_Pin, int CW_Pin, int CCW_Pin, float ramp_scale , int pwm_max, int mid_pos) {

  _PWM_Pin = PWM_Pin;
  _CW_Pin  = CW_Pin;
  _CCW_Pin = CCW_Pin;

  pinMode(_PWM_Pin, OUTPUT);
  pinMode(_CW_Pin,  OUTPUT);
  pinMode(_CCW_Pin, OUTPUT);

  _ramp_scale = ramp_scale;
  _pwm_max    = pwm_max;
  _mid_pos    = mid_pos;

}
int MOTOR_CONTROL::PWM_from_Ramp(int ramp_in, int scale , int pwm_max) {

  int val = 0;
  if (ramp_in < 0) {
    (val = -1 * ramp_in);
  }
  if (ramp_in >= 0) {
    (val = 1 * ramp_in);
  }

  val = val * scale;
  if (val > pwm_max)(val = pwm_max);

  return val;
}
void MOTOR_CONTROL::PWM(int pwm_val) {

  int valueMax = 255;

  if (pwm_val < 0) {
    (pwm_val = -1 * pwm_val);

  }
  if (pwm_val >= 0) {
    (pwm_val = 1 * pwm_val);

  }

  int val = pwm_val;

#ifdef ESP32
  valueMax = 255;
  uint32_t duty = (8191 / valueMax) * min(val, valueMax);
  ledcWrite(_PWM_CHannel, duty);
#endif

#ifdef ESP8266
  uint32_t duty = val;
  analogWrite(_PWM_Pin, val);
#endif

  _curr_pwm_val = duty;
}
void MOTOR_CONTROL::GO(int ramp_in, int offset) {

  int val = 0;
  val = ramp_in;
  if (ramp_in < 0) {
    //(val = -1 * ramp_in);
    digitalWrite(_CCW_Pin, LOW);
    digitalWrite(_CW_Pin, HIGH);
  }
  if (ramp_in >= 0) {
    //(val = 1 * ramp_in);
    digitalWrite(_CW_Pin, LOW);
    digitalWrite(_CCW_Pin, HIGH);
  }

  if (val < 0) {
    (val = -1 * val);
  }
  if (ramp_in >= 0) {
    (val = 1 * val);
  }

  _curr_pwm_val = val;

#ifdef ESP32
  int valueMax = 255;
  uint32_t duty = (8191 / valueMax) * min(val, valueMax);
  ledcWrite(_PWM_CHannel, duty);
#endif

#ifdef ESP8266
  analogWrite(_PWM_Pin, val);
#endif

}
float MOTOR_CONTROL::Get_Cur_PWMval(void) {
  return _curr_pwm_val;
}

MOTOR_CONTROL Motor1, Motor2;

// ##########################################################################

void CH2_calcSignal() {
  RC2.calcSignal();
}
void CH3_calcSignal() {
  RC3.calcSignal();
}
void CH1_calcSignal() {
  RC1.calcSignal();
}
void setup()
{
  pinMode(M1_PWM_Pin, OUTPUT);
  pinMode(M2_PWM_Pin, OUTPUT);

  int  timer_range = 1494;
  RC1.isr_setup(CHANNEL_1_PIN, timer_range);
  RC2.isr_setup(CHANNEL_2_PIN, timer_range);
  RC3.isr_setup(CHANNEL_3_PIN, timer_range);

  attachInterrupt(CHANNEL_1_PIN, CH1_calcSignal, CHANGE);
  attachInterrupt(CHANNEL_2_PIN, CH2_calcSignal, CHANGE);
  attachInterrupt(CHANNEL_3_PIN, CH3_calcSignal, CHANGE);

#ifdef ESP32
  float PWM_Scale = 2.5;
  int   PWM_Max = 255;
  int   Mid_pos = 5; // in % von
#endif

#ifdef ESP8266
  float PWM_Scale = 10.23;
  int   PWM_Max = 1023;
  int   Mid_pos = 5; // in % von max
#endif
  Motor1.setup(M1_PWM_Pin, M1_CW_Pin, M1_CCW_Pin, PWM_Scale, PWM_Max, Mid_pos);
  Motor2.setup(M2_PWM_Pin, M2_CW_Pin, M2_CCW_Pin, PWM_Scale, PWM_Max, Mid_pos);
#ifdef ESP32
  // ESP32 needs special function (Channel 0-12 + PWM_Pin) to write PWM
  Motor1.setup_pwm_ESP32(M1_PWM_Pin, M1_PWM_Channel, LEDC_TIMER_13_BIT, LEDC_BASE_FREQ);
  Motor2.setup_pwm_ESP32(M2_PWM_Pin, M2_PWM_Channel, LEDC_TIMER_13_BIT, LEDC_BASE_FREQ);
#endif
  Serial.begin(115200);
}
void loop()
{
  int scale = 1.5;
  int  CH1_norm_puls = RC1.ramp_signal();
  int  CH2_norm_puls = RC2.ramp_signal();
  int  CH3_norm_puls = RC3.ramp_signal();

  Motor1.GO(RC3.ramp_signal(), 0); // ( Speed , lenken) // Steuer Hebel RC3.ramp_signal()
  Motor2.GO(RC3.ramp_signal(), 0);

  Serial.print(RC2.ramp_signal(), DEC);  // print as an ASCII-encoded decimal
  Serial.print("\t");    // prints a tab

  Serial.print(RC3.ramp_signal(), DEC);  // print as an ASCII-encoded decimal
  Serial.print("\t");    // prints a tab

  /*
    Serial.print(RC3.ramp_signal() , DEC); // print as an ASCII-encoded decimal
    Serial.print("\t");    // prints a tab

    Serial.print(RC1.get_pulse_time(), DEC);  // print as an ASCII-encoded decimal
    Serial.print("\t");    // prints a tab
    Serial.print(RC2.get_pulse_time(), DEC);  // print as an ASCII-encoded decimal
    Serial.print("\t");    // prints a tab
    Serial.print(RC3.get_pulse_time(), DEC);  // print as an ASCII-encoded decimal
    Serial.print("\t");    // prints a tab

    Serial.print(RC3.get_cur_rc_value(), DEC);  // print as an ASCII-encoded decimal
    Serial.print("\t");    // prints a tab
    Serial.print(RC2.get_cur_rc_value(), DEC);  // print as an ASCII-encoded decimal
    Serial.print("\t");    // prints a tab

    Serial.print((int)Motor1.Get_Cur_PWMval(), DEC);  // print as an ASCII-encoded decimal
    Serial.print("\t");    // prints a tab
    Serial.print((int)Motor2.Get_Cur_PWMval(), DEC);  // print as an ASCII-encoded decimal
    Serial.print("\t");    // prints a tab
    Serial.print(CH2_norm_puls, DEC);  // print as an ASCII-encoded decimal
  */
  Serial.print("\n");    // prints a tab
  
  delay(2);
}
